
(module r5rs-tests mzscheme
  (require (lib "match.ss")
           (lib "list.ss")
           (lib "etc.ss")
           (planet "reduction-semantics.ss" ("robby" "redex.plt" 2 4))
           (prefix srfi1: (lib "1.ss" "srfi"))
           "r5rs.scm"
           "app-tests.scm"
           "dw-tests.scm"
           "mv-tests.scm"
           "eval-tests.scm"
           "tl-tests.scm"
           "test.scm")
  
  ;; ============================================================
  ;; TESTING APPARATUS
  
  (define-struct r5test (test expected proof-reds?))

  (define (make-r5test/v t expected) 
    (make-r5test `(store () (dw () ,t)) 
                 (list `(store () (dw () (#%values ,expected))))
                 #t))
  (define (make-r5test/e t err)
    (make-r5test `(store () (dw () ,t))
                 (list `(error ,err))
                 #t))
  
  (define (run-a-test test verbose?)
    (unless verbose?
      (printf ".") 
      (flush-output))
    (let ([t (r5test-test test)]
          [expected (r5test-expected test)])
      (set! test-count (+ test-count 1))
      (when verbose? (printf "testing ~s ... " t))
      (flush-output)
      (with-handlers ([exn:fail:duplicate?
                       (lambda (e)
                         (set! failed-tests (+ failed-tests 1))
                         (unless verbose? 
                           (printf "\ntesting ~s ... " t))
                         (raise e))])
        (let* ([results (evaluate (let ([reds
                                         (if (r5test-proof-reds? test)
                                             reductions
                                             non-proof-reductions)])
                                    (λ (term) 
                                      (reduce reds term)))
                                  t
                                  (or verbose? 'dots)
                                  verify-p*)]
               [rewritten-results (remove-duplicates (map rewrite-actual results))])
          (for-each (verify-a* t) results)
	  (unless (set-same? expected rewritten-results equal?)
	    (set! failed-tests (+ failed-tests 1))
            (unless verbose? 
              (printf "\ntesting ~s ... " t))
	    (printf "TEST FAILED!~nexpected:~a\nrewritten-received:~a\nreceived:~a\n\n"
		    (combine-in-lines expected)
		    (combine-in-lines rewritten-results)
                    (combine-in-lines results)))))))
  
  (define p*-pattern (compile-pattern lang 'p*))
  (define a*-pattern (compile-pattern lang 'a*))
  (define error-pattern (compile-pattern lang '(error string)))
  (define verified-terms 0)
  (define (verify-p* sexp)
    (let ([m (or (match-pattern p*-pattern sexp)
                 (match-pattern error-pattern sexp))])
      (unless m
        (newline)
        (error 'verify-p* "didn't match ~s" sexp))
      (set! verified-terms (+ verified-terms 1))))
  
  (define ((verify-a* orig-sexp) sexp)
    (unless (match-pattern a*-pattern sexp)
      (newline) 
      (error 'verify-a* "didn't match ~s\noriginal term ~s" sexp orig-sexp)))
  
  (define (remove-duplicates lst)
    (let ([ht (make-hash-table 'equal)])
      (for-each (λ (x) (hash-table-put! ht x #t)) lst)
      (hash-table-map ht (λ (x y) x))))
  
  (define (combine-in-lines strs) (apply string-append (map (λ (x) (format "\n  ~s" x)) strs)))
  
  (define (rewrite-actual actual)
    (match actual
      [`(error ,str) actual]
      [`(store ,@(xs ...)) 
        (let loop ([actual actual])
          (subst-:-vars actual))]
      [_
       (error 'rewrite-actual "unknown actual ~s\n" actual)]))
  
  (define (subst-:-vars exp)
    (match exp
      [`(store ,str (dw ,dw ,exp))
        (let* ([sub-thru? (λ (x) (regexp-match #rx":" (format "~a" (car x))))]
               [sub-vars (filter sub-thru? str)])
          ;; have to substitute to a fixed point
          (let loop ([term `(store ,(filter (λ (x) (not (sub-thru? x))) str) (dw ,dw ,exp))])
            (let ([next (do-one-subst sub-vars term)])
              (cond 
                [(equal? term next) next]
                [else (loop next)]))))]
      [`(error ,string) string]
      [_ (error 'subst-:-vars "unknown exp ~s" exp)]))
  
  
  (define (do-one-subst sub-vars term)
    (match term
      [`(store ,str (dw ,dw ,exp))
        (let* ([keep-vars 
                (map (λ (pr)
                       `(,(car pr)
                          ,(r5rs-subst-all (map car sub-vars)
                                           (map cadr sub-vars)
                                           (cadr pr))))
                     str)])
          `(store ,keep-vars (dw ,dw ,(r5rs-subst-all (map car sub-vars) (map cadr sub-vars) exp))))]))

  
  (define test-count 0)
  (define failed-tests 0)
  
  (define r5-specific-tests
    (list
     (make-r5test/v '(#%+) 0)
     (make-r5test/v '(#%+ 1) 1)
     (make-r5test/v '(#%+ 1 2) 3)
     (make-r5test/v '(#%+ 1 2 3) 6)
     
     (make-r5test/v '(#%- 1) -1)
     (make-r5test/v '(#%- 1 2) -1)
     (make-r5test/v '(#%- 1 2 3) -4)
     
     (make-r5test/v '(#%*) 1)
     (make-r5test/v '(#%* 2) 2)
     (make-r5test/v '(#%* 2 3) 6)
     (make-r5test/v '(#%* 2 3 4) 24)
     
     (make-r5test/v '(#%/ 2) 1/2)
     (make-r5test/v '(#%/ 1 2) 1/2)
     (make-r5test/v '(#%/ 1 2 3) 1/6)
     
     (make-r5test/e '(#%/ #f) "arith-op applied to non-number, arg 1")
     
     (make-r5test/e '(#%/ 1 2 3 4 5 0 6) "divison by zero")
     (make-r5test/e '(#%/ 0) "divison by zero")
     
     (make-r5test '(store () (dw () ((lambda (x) (#%+ x x)) #f))) 
                  (list '(error "arith-op applied to non-number, arg 1")
                        '(error "arith-op applied to non-number, arg 2"))
                  #t)
     
     (make-r5test/v '(#%car ((lambda (x) (#%cons x #%null)) 3)) 3)
     (make-r5test/v '((lambda (x) x) 3) 3)
     (make-r5test/v '((lambda (x y) (#%- x y)) 6 5) 1)
     (make-r5test/e '((lambda () (#%+ x y z)) 3 4 5)
                    "arity mismatch")
     (make-r5test/v '((lambda (x y z) (#%+ x y z)) 3 4 5) 12)
     (make-r5test/v '((lambda (x y) (#%+ x y)) (#%+ 1 2) (#%+ 3 4)) 10)
     (make-r5test/v '((lambda (x1 x2 dot y) (#%car y)) 1 2 3 4) 3)
     (make-r5test/v '((lambda (x dot y) (#%car y)) 1 2 3 4) 2)
     (make-r5test/v '((lambda (x dot y) x) 1) 1)
     (make-r5test/e '((lambda (x y dot z) x) 1)
                    "arity mismatch")
     (make-r5test/v '((lambda args (#%car (#%cdr args))) 1 2 3 4 5 6) 2)
     (make-r5test/v '((lambda args (#%eqv? args args)) 1 2) #t)
     (make-r5test/v '((lambda args ((lambda (y) args) (set! args 50)))) 50)
     (make-r5test/v '(if ((lambda (x) x) 74) ((lambda () 6)) (6 54)) 6)
     (make-r5test/e '(1 1) "can't apply non-function")
     (make-r5test/e '(if ((lambda (x) x) #f) ((lambda () 6)) (6 54))
                    "can't apply non-function")
  
     (make-r5test/e '((lambda (x y) x) (lambda (x) x))
                    "arity mismatch")
     
     (make-r5test/v '(if #t 12) 12)
     (make-r5test/v '(if #f 12) 'unspecified)
     (make-r5test/v '(begin (if #f 12) 14) 14)
     (make-r5test/v '((lambda (x) (if #t (set! x 45)) x) 1) 45)
     (make-r5test/v '((lambda (x) (if #f (set! x 45)) x) 1) 1)
     
     (make-r5test/v '(#%call/cc (lambda (k) (#%cons 1 (#%cons 2 (#%cons 3 (k 5)))))) 5)
     (make-r5test/v '(#%call-with-values (lambda () (#%call/cc (lambda (k) (k)))) #%+) 0)
     (make-r5test/v '(#%call-with-values (lambda () (#%call/cc (lambda (k) (k 1 2)))) #%+) 3)
     (make-r5test/v '((#%call/cc #%values) #%values) '#%values)
     (make-r5test '(store () (dw () (define x ((lambda () (#%values 1 2 3))))))
                  (list '(error "context received wrong # of values"))
                  #t)
     (make-r5test '(store ()
                          (dw ()
                              (define x 0)
                              (define f
                                (lambda ()
                                  (set! x (#%+ x 1))
                                  (#%values x x)))
                              (#%call-with-values f (lambda (x y) x))
                              (#%call-with-values f (lambda (x y) x))))
                  (list
                   '(store ((x 2)
                            (f (lambda ()
                                 (set! x (#%+ x 1))
                                 (#%values x x))))
                           (dw () (#%values 2))))
                  #t)
     (make-r5test/v '((lambda (x) (#%call-with-values x (lambda (x y) x)))
                      (lambda () (#%values (#%+ 1 2) 2)))
                    3)
     
     (make-r5test/v '((if #t #%call-with-values) (lambda () (#%+ 1 1)) (lambda (x) x))
                    2)
     
     (make-r5test/v '(#%call-with-values (lambda () (#%values (#%+ 1 2) (#%+ 2 3))) #%+) 8)
     (make-r5test/v '(#%call-with-values #%* #%+) 1)
     (make-r5test/v '(#%call-with-values (lambda () (#%apply #%values (#%cons 1 (#%cons 2 #%null)))) #%+) 3)
     (make-r5test/v '(#%call-with-values (lambda () 1) #%+) 1)

     (make-r5test/e
      '(#%call-with-values (lambda ()
                             ((lambda (f) 
                                (f ((lambda (id) id) (lambda (x) (x x)))
                                   (lambda (x y) x)))
                              #%values))
                           (lambda (a b) (a b)))
      "arity mismatch")
     
     
     (make-r5test/v '(if (#%null? (#%cons 1 (#%cons 2 (#%cons (lambda (x) x) #%null)))) 0 1) 
                    1)
     (make-r5test/v '(#%null? (#%cons 1 (#%cons 2 (#%cons (lambda (x) x) #%null)))) #f)
     (make-r5test/v '(#%null? (#%cons 1 2)) #f)
     (make-r5test/v '(#%null? #%null) #t)
     (make-r5test/v '(#%pair? #%null) #f)
     (make-r5test/v '(#%pair? (#%cons 1 1)) #t)
     (make-r5test/v '(#%null? (#%list 1 2)) #f)
     (make-r5test/v '(#%pair? (#%list 1)) #t)
     (make-r5test/v '(#%pair? (#%list)) #f)
     (make-r5test/v '(#%null? (#%list)) #t)
     
     (make-r5test/v 
      '(#%eval (#%list '#%+ (#%list '#%+ 1 2) 1)) 4)
     
     (make-r5test/v '(#%eval (#%list (lambda (x) x) 1)) 1)
     (make-r5test/v '(#%eval (#%list #%null? 1)) #f)
     (make-r5test/v '((lambda (f)
                        (set! f (lambda (x) x))
                        (#%eqv? (#%eval f) f))
                      1)
                    #t)
     
     (make-r5test/e '(#%eval (#%cons 'a 'b))
                    "malformed expression: (a dot b)")
     
     (make-r5test/v ''#f #f)
     (make-r5test/v ''#t #t)
     (make-r5test/v ''1 1)
     (make-r5test/v ''x ''x)
     (make-r5test/v ''#%null ''#%null)
     (make-r5test/v '(#%null? '#%null) #f)
     (make-r5test/v ''unspecified ''unspecified)
     
     (make-r5test/v (let ([Y '(lambda (le)
                                ((lambda (f) (f f))
                                 (lambda (f)
                                   (le (lambda (z) ((f f) z))))))])
                      `((,Y
                          (lambda (length)
                            (lambda (l)
                              (if (#%null? l)
                                  0
                                  (#%+ (length (#%cdr l)) 1)))))
                        (#%cons 1 #%null)))
                    1)
     
     (make-r5test/v '((lambda (x) ((lambda (y) x) (set! x 5))) 3) 5)
     (make-r5test '(store 
                    ()
                    (dw
                     ()
                     (((lambda (a b ret) ((lambda (x y) ret) (set! ret a) (set! ret b)))
                       (lambda () 1)
                       (lambda () 3)
                       5))))
                  '((store () (dw () (#%values 1)))
                    (store () (dw () (#%values 3))))
                  #t)
     (make-r5test/v '((lambda (x) ((lambda (y) (#%car (#%cdr x))) (#%set-car! (#%cdr x) 400))) 
                      (#%cons 1 (#%cons 2 #%null)))
                    400)
     (make-r5test/v '((lambda (x) ((lambda (y) (#%cdr (#%cdr x))) (#%set-cdr! (#%cdr x) 400)))
                      (#%cons 1 (#%cons 2 #%null))) 
                    400)
     
     
     (make-r5test/v '((lambda (x y) (#%+ x y)) ((lambda (x) x) 3) ((lambda (x) x) 4)) 
                    7)
     (make-r5test/v '((lambda (x) ((lambda (a b) x) (set! x (#%- x)) (set! x (#%- x)))) 1) 
                    1)
     (make-r5test/v '(#%eqv? #t #t) #t)
     (make-r5test/v '(#%eqv? #t #f) #f)
     (make-r5test '(store () (dw () (#%eqv? (lambda (x) x) (lambda (x) x))))
                  (list '(store () (dw () (#%values #f)))
                        '(store () (dw () (#%values #t))))
                  #t)
     (make-r5test '(store () (dw () (#%eqv? (lambda (x) x) (lambda (x) x))))
                  (list '(store () (dw () (#%values #f))))
                  #f)
     (make-r5test/v '(#%eqv? (#%cons 1 2) (#%cons 1 2)) #f)
     (make-r5test/v '((lambda (x) (#%eqv? x x)) (#%cons 1 2)) #t)
     (make-r5test/v '((lambda (x) (#%eqv? x x)) (lambda (x) x)) #t)
     (make-r5test/v '((lambda (x) (begin (set! x 5) (set! x 4) (set! x 3) x)) 0) 3)
     (make-r5test/v '((lambda (x y) (x y)) #%+ 0) 0)
     (make-r5test/v '(#%apply #%+ (#%cons 1 (#%cons 2 #%null))) 3)
     (make-r5test '(store () (dw () (#%apply #%apply #%values '(()))))
                  (list '(store () (dw () (#%values))))
                  #f)
     
     (make-r5test/v '((lambda (x) x) (#%values 1)) 1)
     (make-r5test '(store () (dw () (#%values 1 2)))
                  (list '(store () (dw () (#%values 1 2))))
                  #t)
     (make-r5test '(store () (dw () ((lambda (x) (#%values x x x)) 1) 1)) 
                  (list '(store () (dw () (#%values 1))))
                  #t)
     (make-r5test/v '((lambda args (#%apply #%+ args)) 1 2 3 4) 10)
     (make-r5test/v '((lambda (f) (#%eqv? (f 1) (f 1))) (lambda args (#%car args))) #t)
          
     (make-r5test/v '(begin (#%values) 1) 1)
     (make-r5test/v '(#%+ 1 (begin (#%values 1 2 3) 1)) 2)
     
     (make-r5test '(store () (dw () (#%eval '(define x 1)) x))
                  (list '(error "eval only takes expressions"))
                  #t)
    
     (make-r5test '(store () 
                          (dw ()
                              (define length
                                (lambda (l)
                                  (if (#%null? l)
                                      0
                                      (#%+ 1 (length (#%cdr l))))))
                              (length (#%list 1 2 3))))
                  (list '(store ((length
                                  (lambda (l)
                                    (if (#%null? l)
                                        0
                                        (#%+ 1 (length (#%cdr l)))))))
                                (dw ()
                                    (#%values 3))))
                  #t)
     
     (make-r5test '(store () (dw () (beginD)))
                  (list '(store () (dw () (#%values unspecified))))
                  #f)
     
     (make-r5test '(store () (dw () (beginD) (beginD) (beginD)))
                  (list '(store () (dw () (#%values unspecified))))
                  #f)
     
     (make-r5test '(store () (dw () (beginD 1 2 (beginD 3 4)) (beginD 1 2 (beginD 3 4)) (beginD 1 2 (beginD 3 4))))
                  (list '(store () (dw () (#%values 4))))
                  #f)
     
     ;; testing implicit begin & substitution
     (make-r5test/v '(((lambda (x) x (lambda () x x)) 1)) 1)
     
     ;; sexp contexts for the top-level
     (make-r5test '(store () (dw () (beginD '(define x '1) (define x '2) x)))
                  (list '(store ((x 2)) (dw () (#%values 2))))
                  #t)
     
     ;; test non-determinism in spec (a single application can go two different ways
     ;; at two different times)
     (make-r5test '(store () (dw () 
                                 (define x #%null)
                                 (define twice (lambda (f) (f) (f)))
                                 (twice
                                  (lambda ()
                                    ((lambda (p q) 1)
                                     (set! x (#%cons 1 x))
                                     (set! x (#%cons 2 x)))))))
                  (list
                   '(store ((x (#%cons 1 (#%cons 2 (#%cons 1 (#%cons 2 #%null)))))
                            (twice (lambda (f) (f) (f))))
                           (dw () (#%values 1)))
                   '(store ((x (#%cons 2 (#%cons 1 (#%cons 1 (#%cons 2 #%null)))))
                            (twice (lambda (f) (f) (f))))
                           (dw () (#%values 1)))
                   '(store ((x (#%cons 1 (#%cons 2 (#%cons 2 (#%cons 1 #%null)))))
                            (twice (lambda (f) (f) (f))))
                           (dw () (#%values 1)))
                   '(store ((x (#%cons 2 (#%cons 1 (#%cons 2 (#%cons 1 #%null)))))
                            (twice (lambda (f) (f) (f))))
                           (dw () (#%values 1))))
                  #t)

     
     ;; dynamic wind and multiple values
     (make-r5test '(store () (dw () (#%dynamic-wind #%values (lambda () (#%values 1 2)) #%values)))
                  (list '(store () (dw () (#%values 1 2))))
                  #t)
     
     ;; dynamic-wind given non-lambda procedure values
     (make-r5test '(store () (dw () (#%dynamic-wind #%values #%values #%values)))
                  (list '(store () (dw () (#%values))))
                  #t)
     (make-r5test '(store () (dw () (#%dynamic-wind #%values (lambda x x) #%values)))
                  (list '(store () (dw () (#%values #%null))))
                  #t)

     
     (make-r5test/e '(#%dynamic-wind 1 1 1) "dynamic-wind expects arity 0 procs")

     ;; make sure that dynamic wind signals non-proc errors directly
     ;; instead of calling procedures
     (make-r5test/e '(#%dynamic-wind (lambda () (#%car 1)) 1 2)
                    "dynamic-wind expects arity 0 procs")
     (make-r5test/e '(#%dynamic-wind (lambda () (#%car 1)) (lambda (x) x) (lambda (y) y))
                    "dynamic-wind expects arity 0 procs")
     
     (make-r5test/e '(#%dynamic-wind (lambda () (#%car 1)) (lambda (x dot y) x) (lambda () 1))
                    "dynamic-wind expects arity 0 procs")
     (make-r5test/v '(#%dynamic-wind #%+ (lambda y 2) #%*)
                    2)
     (make-r5test/v '(#%dynamic-wind #%values #%list (lambda y y))
                    '#%null)

     (make-r5test '(store () (dw ()
                                 (define x 1)
                                 ((lambda (t) (t) (t))
                                  (lambda ()
                                    ((lambda (a b) x)
                                     (set! x (#%+ x 1))
                                     (set! x (#%* x 2)))))))
                  (list '(store ((x 7)) (dw () (#%values 7)))
                        '(store ((x 8)) (dw () (#%values 8)))
                        '(store ((x 9)) (dw () (#%values 9)))
                        '(store ((x 10)) (dw () (#%values 10))))
                  #t)
     
     ;; arity of various primitives
     
     (make-r5test/e '(#%call-with-values (lambda (x) x) (lambda (y) y))
                    "arity mismatch")
     (make-r5test/e '(#%/) "arity mismatch")
     (make-r5test/e '(#%-) "arity mismatch")
     (make-r5test/e '(#%cons) "arity mismatch")
     (make-r5test/e '(#%null?) "arity mismatch")
     (make-r5test/e '(#%pair?) "arity mismatch")
     (make-r5test/e '(#%car) "arity mismatch")
     (make-r5test/e '(#%cdr) "arity mismatch")
     (make-r5test/e '(#%set-car!) "arity mismatch")
     (make-r5test/e '(#%set-cdr!) "arity mismatch")
     (make-r5test/e '(#%call/cc) "arity mismatch")
     (make-r5test/e '(#%eqv?) "arity mismatch")
     (make-r5test/e '(#%apply) "arity mismatch")
     (make-r5test/e '(#%apply #%values) "arity mismatch")
     (make-r5test/e '(#%call-with-values) "arity mismatch")
     (make-r5test/e '(#%eval) "arity mismatch")
     
     (make-r5test/e '(#%dynamic-wind 1) "arity mismatch")
     
     (make-r5test/e '(#%apply 1 2) "can't apply non-function")
     (make-r5test/e '(#%apply 1 #%null) "can't apply non-function")
     (make-r5test/e '(#%apply #%values 2) "apply's last argument non-list")
     (make-r5test/e '(#%car 1) "can't take car of non-pair")
     (make-r5test/e '(#%cdr 1) "can't take cdr of non-pair")
     (make-r5test/e '(#%set-car! 2 1) "can't set-car! on a non-pair")
     (make-r5test/e '(#%set-cdr! 1 2) "can't set-cdr! on a non-pair")
     
     (make-r5test/e '(#%call/cc 1) "can't apply non-function")
     (make-r5test/e '(#%call-with-values 1 2) "can't apply non-function")
     
     (make-r5test/v '((lambda (x) (#%eqv? 'x 1)) 1) #f)
     (make-r5test/v '(#%eqv? 'x ((lambda (x) 'x) 2)) #t)

     (make-r5test/v ''(a dot x) '(#%cons 'a 'x))
     (make-r5test/v ''(a b c d dot x) '(#%cons 'a (#%cons 'b (#%cons 'c (#%cons 'd 'x)))))
     (make-r5test/v ''(2 dot 1) '(#%cons 2 1))
     (make-r5test/v '((#%eval '(lambda x (#%car x))) 1) 1)
     
     ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
     ;;
     ;; tests from R5RS
     ;;
     
     ; ----
     ; 4.1.1
     (make-r5test '(store () 
                          (dw ()
                              (define x 28)
                              x))
                  (list '(store ((x 28)) (dw () (#%values 28))))
                  #t)

     
     ; ----
     ; 4.1.3
     (make-r5test/v '(#%+ 3 4) 7)
     (make-r5test/v '((if #f #%+ #%*) 3 4) 12)


     ; ----
     ; 4.1.4
     (make-r5test/v '(lambda (x) (#%+ x x)) '(lambda (x) (#%+ x x)))
     (make-r5test/v '((lambda (x) (#%+ x x)) 4) 8)

     (make-r5test '(store ()
                          (dw ()
                              (define reverse-subtract
                                (lambda (x y) (#%- y x)))
                              (reverse-subtract 7 10)))
                  (list
                   '(store ((reverse-subtract
                             (lambda (x y) (#%- y x))))
                           (dw ()
                               (#%values 3))))
                  #t)
     
     (make-r5test '(store ()
                          (dw ()
                              (define add4
                                ((lambda (x)
                                   (lambda (y)
                                     (#%+ x y)))
                                 4))
                              (add4 6)))
                  (list
                   '(store ((add4 (lambda (y) (#%+ 4 y))))
                           (dw ()
                               (#%values 10))))
                  #t)
     
     (make-r5test/v '((lambda x x) 3 4 5 6)
                    '(#%cons 3 (#%cons 4 (#%cons 5 (#%cons 6 #%null)))))
     (make-r5test/v '((lambda (x y dot z) z) 3 4 5 6)
                    '(#%cons 5 (#%cons 6 #%null)))
     

; ----
; 4.1.5

; (if (> 3 2) 'yes 'no)                   ===>  yes
; (if (> 2 3) 'yes 'no)                   ===>  no

     ; (make-r5test/v '(if (> 3 2) (#%- 3 2) (#%+ 3 2)) 1)

; ----
; 4.1.6

     (make-r5test '(store () (dw ()
                                 (define x 2)
                                 (#%+ x 1)))
                  (list
                   '(store ((x 2)) 
                           (dw () 
                               (#%values 3))))
                  #t)
     
     (make-r5test '(store () (dw ()
                                 (define x 2)
                                 (#%+ x 1) 
                                 (set! x 4)))
                  (list
                   '(store ((x 4)) 
                           (dw () 
                               (#%values unspecified))))
                  #t)
     
     (make-r5test '(store () (dw ()
                                 (define x 2)
                                 (#%+ x 1) 
                                 (set! x 4)
                                 (#%+ x 1)))
                  (list
                   '(store ((x 4)) 
                           (dw () 
                               (#%values 5))))
                  #t)


; ----
; 4.2.2 
     
     (make-r5test '(store
                    ()
                    (dw
                     ()
                     (define even?
                       (lambda (n)
                         (if (#%eqv? 0 n)
                             #t
                             (odd? (#%- n 1)))))
                     (define odd?
                       (lambda (n)
                         (if (#%eqv? 0 n)
                             #f
                             (even? (#%- n 1)))))
                     ;; using 88 here runs, but isn't really much more useful
                     ;; for testing purposes (it also takes > 1000 reductions)
                     (even? 2)))
                  (list
                   '(store
                     ((even?
                       (lambda (n)
                         (if (#%eqv? 0 n)
                             #t
                             (odd? (#%- n 1)))))
                      (odd?
                       (lambda (n)
                         (if (#%eqv? 0 n)
                             #f
                             (even? (#%- n 1))))))
                     (dw ()
                         (#%values #t))))
                  #t)
                   

; ----
; 4.2.3
     (make-r5test '(store () (dw () (define x 0) (begin (set! x 5) (#%+ x 1))))
                  (list '(store ((x 5)) (dw () (#%values 6))))
                  #t)


; ----
; 5.2.1

     (make-r5test '(store () (dw () (define add3 (lambda (x) (#%+ x 3))) (add3 3)))
                  (list '(store ((add3 (lambda (x) (#%+ x 3)))) (dw () (#%values 6))))
		  #t)
     (make-r5test '(store () (dw () (define first #%car) (first '(1 2))))
                  (list '(store ((first #%car)) (dw () (#%values 1))))
		  #t)


; ----
; 6.1

     (make-r5test/v '(#%eqv? 'a 'a) #t)
     (make-r5test/v '(#%eqv? 'a 'b) #f)
     (make-r5test/v '(#%eqv? 2 2) #t)
     (make-r5test/v '(#%eqv? '() '()) #t)
     (make-r5test/v '(#%eqv? 100000000 100000000) #t)
     (make-r5test/v '(#%eqv? (#%cons 1 2) (#%cons 1 2)) #f)
     (make-r5test '(store () (dw () (#%eqv? (lambda () 1) (lambda () 2))))
                  (list '(store () (dw () (#%values #f))))
                  #f)
     (make-r5test/v '(#%eqv? #f 'nil) #f)
     (make-r5test/v '(#%eqv? #f '()) #f)
     (make-r5test/v '((lambda (p) (#%eqv? p p)) (lambda (x) x)) #t)

     (make-r5test 
      '(store () (dw ()
                     (define gen-counter
                       (lambda ()
                         ((lambda (n) 
                            (lambda () (set! n (#%+ n 1)) n))
                          0)))
                     ((lambda (g) (#%eqv? g g))
                      (gen-counter))))
      (list '(store ((gen-counter
                      (lambda ()
                        ((lambda (n) 
                           (lambda () (set! n (#%+ n 1)) n))
                         0))))
                    (dw () 
                        (#%values #t))))
      #f)
     
     (make-r5test 
      '(store () (dw ()
                     (define gen-counter
                       (lambda ()
                         ((lambda (n) 
                            (lambda () (set! n (#%+ n 1)) n))
                          0)))
                     (#%eqv? (gen-counter) (gen-counter))))
      (list '(store ((gen-counter
                      (lambda ()
                        ((lambda (n) 
                           (lambda () (set! n (#%+ n 1)) n))
                         0))))
                    (dw () 
                        (#%values #f))))
      #f)
     
     
      
; ----
; 6.3.2

     (make-r5test '(store ()
                          (dw ()
                              (define x (#%list 'a 'b 'c))
                              (define y x)
                              y))
                  (list
                   '(store ((x (#%cons 'a (#%cons 'b (#%cons 'c #%null))))
                            (y (#%cons 'a (#%cons 'b (#%cons 'c #%null)))))
                           (dw ()
                               (#%values (#%cons 'a (#%cons 'b (#%cons 'c #%null)))))))
		  #t)
     
     (make-r5test '(store ()
                          (dw ()
                              (define x (#%list 'a 'b 'c))
                              (define y x)
                              (#%set-cdr! x 4)
                              x))
                  (list
                   '(store ((x (#%cons 'a 4))
                            (y (#%cons 'a 4)))
                           (dw ()
                               (#%values (#%cons 'a 4)))))
		  #t)
     
     (make-r5test '(store ()
                          (dw ()
                              (define x (#%list 'a 'b 'c))
                              (define y x)
                              (#%set-cdr! x 4)
                              (#%eqv? x y)))
                  (list
                   '(store ((x (#%cons 'a 4))
                            (y (#%cons 'a 4)))
                           (dw ()
                               (#%values #t))))
		  #t)
     
     (make-r5test '(store ()
                          (dw ()
                              (define x (#%list 'a 'b 'c))
                              (define y x)
                              (#%set-cdr! x 4)
                              y))
                  (list
                   '(store ((x (#%cons 'a 4))
                            (y (#%cons 'a 4)))
                           (dw ()
                               (#%values (#%cons 'a 4)))))
		  #t)

     ; ----
     ; 6.4
     (make-r5test/v '(#%apply #%+ (#%list 3 4)) 7)
    
     (make-r5test
      '(store ()
              (dw ()
                  (define compose
                    (lambda (f g)
                      (lambda args
                        (f (#%apply g args)))))
                  
                  (define sqrt (lambda (x) (if (#%eqv? x 900) 30 #f)))
                  
                  ((compose sqrt #%*) 12 75)))
      (list '(store ((compose (lambda (f g)
                                (lambda args
                                  (f (#%apply g args)))))
                     (sqrt (lambda (x) (if (#%eqv? x 900) 30 #f))))
                    (dw ()
                        (#%values 30))))
      #t)))
  
  (define (conv-base base vec)
    (let loop ([i (vector-length vec)]
               [acc 0])
      (cond
        [(zero? i) acc]
        [else (loop (- i 1)
                    (+ acc (* (expt base (- i 1))
                              (vector-ref vec (- i 1)))))])))
  
  (define (deconv-base base number)
    (list->vector
     (let loop ([i number])
       (cond
         [(zero? i) '()]
         [else (cons (modulo i base)
                     (loop (quotient i base)))]))))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;; tests from simpler systems
  ;;
  
  (define (#%ize exp)
    (let loop ([exp exp])
      (cond
        [(pair? exp) (cons (loop (car exp)) (loop (cdr exp)))]
        [(memq exp names-that-need#%) 
         (string->symbol (format "#%~a" exp))]
        [else exp])))
  
  (define names-that-need#%
    (map (λ (x) 
           (let ([str (symbol->string x)])
             (string->symbol (substring str 2 (string-length str)))))
    '(#%null
      #%car #%cdr #%set-car! #%set-cdr! #%list
      #%call/cc #%dynamic-wind
      #%eqv?
      #%apply
      #%eval
      #%cons #%pair? #%null? #%car #%cdr  ; list functions
      #%set-car! #%set-cdr! #%list
      #%+ #%- #%* #%/                             ; math functions
      #%call/cc #|throw push pop|# #%dynamic-wind  ; call/cc functions
      #%values #%call-with-values              ; values functions
      )))
  
  (define (build-suite-tests test-suite conv-in conv-out)
    (map (λ (test)
           (make-r5test (conv-in (test-input test))
                        (map conv-out (test-expecteds test))
			#t))
         (test-suite-tests test-suite)))
  
  (define (conv-app-in app-exp) 
    (match app-exp
      [`(store ,s ,exp) `(store ,s (dw () ,(#%ize exp)))]
      [else (error 'conv-app-in "unknown ~s" app-exp)]))
  
  (define (conv-app-out app-exp) 
    (match app-exp
      [`(store ,s ,v) `(store ,(filter-number-suffixes-out s) (dw () (#%values ,v)))]
      [else (error 'conv-app-out "unknown ~s" app-exp)]))
  
  ;; store -> store
  ;; removes the elements of the store that come from the app reduction, since they
  ;; will also be removed by the cleanup phase when running the test suite in the 
  ;; full semantics
  (define (filter-number-suffixes-out s)
    (filter (λ (pr) (not (regexp-match #rx"[0-9]$" (symbol->string (car pr)))))
            s))
  
  (define app-tests (build-suite-tests app-ts conv-app-in conv-app-out))
  
  (define (conv-mv-in mv-exp) 
    (match mv-exp
      [`(store ,s ,mv-exp)
        ;; always ignore the store
        `(store () (dw () ,(#%ize mv-exp)))]
      [else (error 'conv-mv-in "unknown ~s" mv-exp)]))
  
  (define (conv-mv-out mv-expected)
    (match mv-expected
      [`(error ,(? (λ (x) (regexp-match #rx"^\\+" x)) str))
        `(error ,(regexp-replace #rx"^." str "arith-op"))]
      [`(error ,mv-expected) 
        `(error ,mv-expected)]
      [`(store ,s ,mv-expected)
        ;; always ignore the store
        `(store () (dw () ,(#%ize mv-expected)))]
      [else
       (error 'conv-mv-out "unknown ~s" mv-expected)]))
  (define mv-tests (build-suite-tests mv-ts conv-mv-in conv-mv-out))
  
  (define (conv-eval-in exp)
    (match exp
      [`(store () ,exp)
        `(store () (dw () ,(#%ize exp)))]))
  (define (conv-eval-out exp) `(store () (dw () (#%values ,exp))))
  (define (remove-last-test lst)
    (cond
      [(null? lst) null]
      [(null? (cdr lst)) null]
      [else (cons (car lst) (remove-last-test (cdr lst)))]))
  (define eval-tests 
    ;; remove the last test, because it uses a different reifier (and so should fail here)
    (remove-last-test (build-suite-tests eval-ts conv-eval-in conv-eval-out)))

  (define (conv-dw-in exp) (#%ize (rewrite-special-thunk exp)))
  
  (define (rewrite-special-thunk exp)
    (match exp
      [`(lambda () (+))
        ;; this simplification saves lots of steps in one of the test cases
        '+]
      [`(,xs ...) (map rewrite-special-thunk exp)]
      [else exp]))
  
  (define (conv-dw-out exp)     
    (match exp
      [`(store ,s (dw ,d ,v))
        (#%ize `(store ,s (dw ,d (#%values ,v))))]))
  
  (define dw-tests (build-suite-tests dw-ts conv-dw-in conv-dw-out))
  
  (define (conv-tl-in exp)
    (let ([exp (convert-inner-begins (#%ize exp))])
      `(store ,(cadr exp) (dw () ,@(cddr exp)))))

  (define (conv-tl-out expected) 
    (convert-inner-begins
     (match expected
       [`(store ,s ,(? string? msg))
         `(error ,msg)]
       [`(store ,s ,v)
         `(store ,s (dw ,() (#%values ,v)))]
       [x (error 'conv-tl-out "~s" x)])))
  
  (define (convert-inner-begins sexp)
    (let loop ([sexp sexp])
      (match sexp
        [`(beginE ,@(xs ...)) `(begin ,@(map loop xs))]
        [(xs ...) (map loop xs)]
        [else sexp])))

  (define tl-tests (build-suite-tests tl-ts conv-tl-in conv-tl-out))

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;; all reduction names are distinct
  ;;
  
  (define (verify-distinct-names)
    (let ([ht (make-hash-table 'equal)])
      (for-each
       (λ (red)
         (hash-table-put! ht (red-name red) (+ 1 (hash-table-get ht (red-name red) 0))))
       reductions)
      (let ([failed? #f])
        (hash-table-for-each
         ht
         (λ (k v)
           (when (v . > . 1)
             (fprintf (current-error-port) "verify-distinct-names: the reduction name ~s appears ~a times\n" k v)
             (set! failed? #t))))
        (when failed?
          (error 'verify-distinct-names)))))
         
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;; all of the tests
  ;;
  
  (define the-sets 
    (list (list "r5" r5-specific-tests)
          (list "app" app-tests)
          (list "mv" mv-tests)
          (list "eval" eval-tests)
          (list "tl" tl-tests)
          (list "dw" dw-tests)))
    
  (define the-tests (apply append (map cadr the-sets)))
  
  (define run-tests
    (opt-lambda ([verbose? #f])
      (verify-distinct-names)
      (time
       (let ()
         (define first? #t)
         (define (run-a-set name set)
           (unless first?
             (if verbose?
                 (printf "\n\n")
                 (printf "\n")))
           (if verbose?
               (printf "~a\n~a tests\n\n" 
                       (apply string (build-list 60 (λ (i) #\-)))
                       name)
               (begin (printf "~a tests " name)
                      (flush-output)))
           (set! first? #f)
           (for-each (λ (x) (run-a-test x verbose?)) set))
         
         (set! failed-tests 0)
         (set! verified-terms 0)
         (for-each (λ (set) (apply run-a-set set)) the-sets)
         (unless verbose? (printf "\n"))
         
         (if (= 0 failed-tests)
             (printf "~a tests, all passed\n" test-count)
             (printf "~a tests, ~a tests failed\n" test-count failed-tests))
         (printf "verified that ~a terms are p* (the terms are not necc. distinct)\n" verified-terms)))
      (collect-garbage) (collect-garbage) (collect-garbage)
      (printf "mem ~s\n" (current-memory-use))))

  (provide run-tests
           the-tests
           
           ;; the test and the expected are not compared with equal?.
           ;; instead, the result of running the test is first simplified
           ;; by substituting all of the variables with a colon in their
           ;; names thru the term, and then the results from the test are
           ;; compared with equal? to the elements of `expected'
           (struct r5test (test            ;; p (from the r5 grammar) [the test]
                           expected))))    ;; (list-of p)

